from greenideas.attributes.attribute_type import AttributeType
from greenideas.attributes.case import Case
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.expansion_spec import ExpansionSpec
from greenideas.rules.grammar_rule import GrammarRule
from greenideas.rules.source_spec import SourceSpec

# PP -> Prep NP
pp__prep_np = GrammarRule(
    SourceSpec(POSType.PP),
    [
        ExpansionSpec(POSType.Prep),
        ExpansionSpec(POSType.NP, {AttributeType.CASE: Case.OBJECTIVE}),
    ],
)

pp_expansions = [pp__prep_np]



from greenideas.attributes.aspect import Aspect
from greenideas.attributes.attribute_type import AttributeType
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.expansion_spec import INHERIT, ExpansionSpec
from greenideas.rules.grammar_rule import GrammarRule
from greenideas.rules.source_spec import SourceSpec

# VP -> Aux_finite VP.participle
# placeholder, need to add additional attributes before implementing this correctly
auxp__auxFinite_vpParticiple = GrammarRule(
    SourceSpec(
        POSType.AuxP,
        {
            AttributeType.ASPECT: [
                Aspect.PERFECT,
            ]
        },
    ),
    [
        ExpansionSpec(
            POSType.Aux_finite,
            {
                AttributeType.ASPECT: INHERIT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.TENSE: INHERIT,
                AttributeType.PERSON: INHERIT,
            },
        ),
        ExpansionSpec(
            POSType.VP,
            {
                AttributeType.ASPECT: INHERIT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.TENSE: INHERIT,
                AttributeType.PERSON: INHERIT,
            },
        ),
    ],
)

auxPerfprog__auxFinite_vpParticiple = GrammarRule(
    SourceSpec(
        POSType.AuxP,
        {
            AttributeType.ASPECT: Aspect.PERFECT_PROGRESSIVE,
        },
    ),
    [
        ExpansionSpec(
            POSType.Aux_finite,
            {
                AttributeType.ASPECT: Aspect.PERFECT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.TENSE: INHERIT,
                AttributeType.PERSON: INHERIT,
            },
        ),
        ExpansionSpec(
            POSType.Be,
            {
                AttributeType.ASPECT: Aspect.PERFECT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.TENSE: INHERIT,
                AttributeType.PERSON: INHERIT,
            },
        ),
        ExpansionSpec(
            POSType.VP,
            {
                AttributeType.ASPECT: INHERIT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.TENSE: INHERIT,
                AttributeType.PERSON: INHERIT,
            },
        ),
    ],
)


auxpProg__auxFinite_vpGerund = GrammarRule(
    SourceSpec(
        POSType.AuxP,
        {
            AttributeType.ASPECT: [Aspect.PROGRESSIVE],
        },
    ),
    [
        ExpansionSpec(
            POSType.Be,
            {
                AttributeType.ASPECT: Aspect.SIMPLE,
                AttributeType.NUMBER: INHERIT,
                AttributeType.TENSE: INHERIT,
                AttributeType.PERSON: INHERIT,
            },
        ),
        ExpansionSpec(
            POSType.VP,
            {
                AttributeType.ASPECT: INHERIT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.TENSE: INHERIT,
                AttributeType.PERSON: INHERIT,
            },
        ),
    ],
)

# AuxP -> Aux_do V_Bare
auxp__auxDo_vpBare = GrammarRule(
    SourceSpec(POSType.AuxP, {AttributeType.ASPECT: Aspect.SIMPLE}),
    [
        ExpansionSpec(
            POSType.Aux_do,
            {
                AttributeType.ASPECT: INHERIT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.TENSE: INHERIT,
                AttributeType.PERSON: INHERIT,
            },
        ),
        ExpansionSpec(POSType.VP_Bare),
    ],
)


auxP_expansions = [
    auxp__auxFinite_vpParticiple,
    auxPerfprog__auxFinite_vpParticiple,
    auxpProg__auxFinite_vpGerund,
    auxp__auxDo_vpBare,
]



from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.expansion_spec import ExpansionSpec
from greenideas.rules.grammar_rule import GrammarRule
from greenideas.rules.source_spec import SourceSpec

# adjP -> Adj
adjP__adj = GrammarRule(
    SourceSpec(POSType.AdjP),
    [
        ExpansionSpec(
            POSType.Adj,
        )
    ],
)

# adjP -> adjP conj adjP
adjP__adjP_conj_adjP = GrammarRule(
    SourceSpec(POSType.AdjP),
    [
        ExpansionSpec(POSType.AdjP),
        ExpansionSpec(POSType.CoordConj),
        ExpansionSpec(POSType.AdjP),
    ],
    weight=0.1,
)

adjP_expansions = [adjP__adj, adjP__adjP_conj_adjP]



from greenideas.exceptions import TwaddleConversionError
from greenideas.parts_of_speech.pos_node import POSNode
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.twaddle.twaddle_formatting_handler import TwaddleFormattingHandler


class TwaddleFormatter:
    def __init__(self):
        self.formatting_handlers: dict[POSType, TwaddleFormattingHandler] = dict()

    def register_formatting_handler(
        self, pos: POSType, handler: TwaddleFormattingHandler
    ):
        self.formatting_handlers[pos] = handler

    def format_node(self, node: POSNode) -> str:
        handler = self.formatting_handlers.get(node.type)
        if handler is None:
            raise TwaddleConversionError(
                f"No formatting handler registered for type {node.type}"
                f"Node has attributes: {node.attributes}"
            )
        return handler.format(node)

    def format(self, tree: POSNode) -> str:
        if not isinstance(tree, POSNode):
            raise TwaddleConversionError("Input must be a POSNode")
        twaddle_string = ""
        if not tree.children:
            twaddle_string = self.format_node(tree)
        else:
            child_tags = [self.format(child) for child in tree.children]
            twaddle_string = " ".join(filter(None, [twaddle_string, *child_tags]))
        return twaddle_string

    def format_as_sentence(self, tree: POSNode) -> str:
        if not isinstance(tree, POSNode):
            raise TwaddleConversionError("Input must be a POSNode")
        twaddle_string = self.format(tree)
        return f"[case:sentence]{twaddle_string}."



from greenideas.attributes.attribute_type import AttributeType
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.expansion_spec import INHERIT, ExpansionSpec
from greenideas.rules.grammar_rule import GrammarRule
from greenideas.rules.source_spec import SourceSpec

# ModalP -> Modal V_AfterModal
modalP__modal_vAfterModal = GrammarRule(
    SourceSpec(POSType.ModalP),
    [
        ExpansionSpec(
            POSType.Modal,
            {
                AttributeType.TENSE: INHERIT,
                AttributeType.ASPECT: INHERIT,
            },
        ),
        ExpansionSpec(
            POSType.VP_AfterModal,
            {
                AttributeType.ASPECT: INHERIT,
            },
        ),
    ],
)

modalP_expansions = [modalP__modal_vAfterModal]



from dataclasses import dataclass, field
from typing import List, Self

from greenideas.attributes.attribute_set import AttributeSet
from greenideas.parts_of_speech.pos_types import POSType


@dataclass
class POSNode:
    type: POSType
    children: List[Self] = field(default_factory=list)
    attributes: AttributeSet = field(default_factory=AttributeSet)

    def __str__(self):
        children = (
            f"[{', '.join(str(child) for child in self.children)}]"
            if self.children
            else None
        )
        return f"{self.type.name}{children if children else ''}"



# S -> NP VP
from greenideas.attributes.aspect import Aspect
from greenideas.attributes.attribute_type import AttributeType
from greenideas.attributes.case import Case
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.expansion_spec import INHERIT, ExpansionSpec
from greenideas.rules.grammar_rule import GrammarRule
from greenideas.rules.source_spec import SourceSpec

s__np_vp = GrammarRule(
    SourceSpec(POSType.S),
    [
        ExpansionSpec(
            POSType.NP,
            {
                AttributeType.ANIMACY: INHERIT,
                AttributeType.CASE: Case.NOMINATIVE,
                AttributeType.NUMBER: INHERIT,
                AttributeType.PERSON: INHERIT,
            },
        ),
        ExpansionSpec(
            POSType.VP,
            {
                AttributeType.ANIMACY: INHERIT,
                AttributeType.ASPECT: Aspect.SIMPLE,
                AttributeType.NUMBER: INHERIT,
                AttributeType.PERSON: INHERIT,
            },
        ),
    ],
)

# S -> NP AuxP
s__np_auxp = GrammarRule(
    SourceSpec(POSType.S),
    [
        ExpansionSpec(
            POSType.NP,
            {
                AttributeType.ANIMACY: INHERIT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.PERSON: INHERIT,
                AttributeType.CASE: Case.NOMINATIVE,
            },
        ),
        ExpansionSpec(
            POSType.AuxP,
            {
                AttributeType.NUMBER: INHERIT,
                AttributeType.PERSON: INHERIT,
                AttributeType.TENSE: INHERIT,
            },
        ),
    ],
)

# S -> NP ModalP
s__np_modalp = GrammarRule(
    SourceSpec(POSType.S),
    [
        ExpansionSpec(
            POSType.NP,
            {
                AttributeType.ANIMACY: INHERIT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.PERSON: INHERIT,
                AttributeType.CASE: Case.NOMINATIVE,
            },
        ),
        ExpansionSpec(
            POSType.ModalP,
            {
                AttributeType.TENSE: INHERIT,
            },
        ),
    ],
)

# S -> S Conj S
s__s_conj_s = GrammarRule(
    SourceSpec(POSType.S),
    [
        ExpansionSpec(POSType.S),
        ExpansionSpec(POSType.CoordConj),
        ExpansionSpec(POSType.S),
    ],
    weight=0.2,
)

s__s_sub_s = GrammarRule(
    SourceSpec(POSType.S),
    [
        ExpansionSpec(POSType.S),
        ExpansionSpec(POSType.Subordinator),
        ExpansionSpec(POSType.S),
    ],
    weight=0.2,
)

s_expansions = [
    s__np_vp,
    s__np_auxp,
    s__np_modalp,
    s__s_conj_s,
    s__s_sub_s,
]



from enum import auto

from greenideas.attributes.grammatical_attribute import GrammaticalAttribute


class Aspect(GrammaticalAttribute):
    SIMPLE = auto()
    PROGRESSIVE = auto()
    PERFECT = auto()
    PERFECT_PROGRESSIVE = auto()

    def __str__(self) -> str:
        return self.name.lower()



from greenideas.attributes.attribute_type import AttributeType
from greenideas.attributes.case import Case
from greenideas.attributes.valency import Valency
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.expansion_spec import INHERIT, ExpansionSpec
from greenideas.rules.grammar_rule import GrammarRule
from greenideas.rules.source_spec import SourceSpec

# VP -> VP AdvP
vp__vp_advp = GrammarRule(
    SourceSpec(POSType.VP),
    [
        ExpansionSpec(
            POSType.VP,
            {
                AttributeType.ASPECT: INHERIT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.TENSE: INHERIT,
                AttributeType.PERSON: INHERIT,
                AttributeType.VALENCY: INHERIT,
            },
        ),
        ExpansionSpec(POSType.AdvP),
    ],
    weight=0.2,
)

# VP1 -> V1
vp1__v = GrammarRule(
    SourceSpec(
        POSType.VP,
        {
            AttributeType.VALENCY: Valency.MONOVALENT,
        },
    ),
    [
        ExpansionSpec(
            POSType.Verb,
            {
                AttributeType.ASPECT: INHERIT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.TENSE: INHERIT,
                AttributeType.PERSON: INHERIT,
                AttributeType.VALENCY: INHERIT,
            },
        ),
    ],
)

# VP2 -> V NP.Acc
vp2__v_npAcc = GrammarRule(
    SourceSpec(POSType.VP, {AttributeType.VALENCY: Valency.DIVALENT}),
    [
        ExpansionSpec(
            POSType.Verb,
            {
                AttributeType.ASPECT: INHERIT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.TENSE: INHERIT,
                AttributeType.PERSON: INHERIT,
                AttributeType.VALENCY: INHERIT,
            },
        ),
        ExpansionSpec(
            POSType.NP,
            {
                AttributeType.NUMBER: INHERIT,
                AttributeType.CASE: Case.OBJECTIVE,
            },
        ),
    ],
)


# VP -> VP PP
vp__vp_pp = GrammarRule(
    SourceSpec(POSType.VP),
    [
        ExpansionSpec(
            POSType.VP,
            {
                AttributeType.ASPECT: INHERIT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.TENSE: INHERIT,
                AttributeType.PERSON: INHERIT,
                AttributeType.VALENCY: INHERIT,
            },
        ),
        ExpansionSpec(POSType.PP),
    ],
    weight=0.2,
)

# VP -> VP Conj VP
vp__vp_conj_vp = GrammarRule(
    SourceSpec(POSType.VP),
    [
        ExpansionSpec(
            POSType.VP,
            {
                AttributeType.ASPECT: INHERIT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.TENSE: INHERIT,
                AttributeType.PERSON: INHERIT,
                AttributeType.VALENCY: INHERIT,
            },
        ),
        ExpansionSpec(POSType.CoordConj),
        ExpansionSpec(
            POSType.VP,
            {
                AttributeType.ASPECT: INHERIT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.TENSE: INHERIT,
                AttributeType.PERSON: INHERIT,
            },
        ),
    ],
    weight=0.2,
)


vp_expansions = [
    vp__vp_advp,
    vp__vp_pp,
    vp__vp_conj_vp,
    vp1__v,
    vp2__v_npAcc,
]



from greenideas.attributes.attribute_type import AttributeType
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.expansion_spec import INHERIT, ExpansionSpec
from greenideas.rules.grammar_rule import GrammarRule
from greenideas.rules.source_spec import SourceSpec

# NP_NoDet -> N
npNodet__n = GrammarRule(
    SourceSpec(POSType.NP_NoDet),
    [
        ExpansionSpec(
            POSType.Noun,
            {
                AttributeType.ANIMACY: INHERIT,
                AttributeType.CASE: INHERIT,
                AttributeType.NUMBER: INHERIT,
            },
        ),
    ],
)

# NP_NoDet -> AdjP NP_NoDet
np_nodet__adjp_np_nodet = GrammarRule(
    SourceSpec(POSType.NP_NoDet),
    [
        ExpansionSpec(POSType.AdjP),
        ExpansionSpec(
            POSType.NP_NoDet,
            {
                AttributeType.ANIMACY: INHERIT,
                AttributeType.CASE: INHERIT,
                AttributeType.NUMBER: INHERIT,
            },
        ),
    ],
    weight=0.2,
)

# NP_NoDet -> AdjP N
npNodet__adjp_n = GrammarRule(
    SourceSpec(POSType.NP_NoDet),
    [
        ExpansionSpec(POSType.AdjP),
        ExpansionSpec(
            POSType.Noun,
            {
                AttributeType.ANIMACY: INHERIT,
                AttributeType.CASE: INHERIT,
                AttributeType.NUMBER: INHERIT,
            },
        ),
    ],
    weight=0.2,
)

# NP_NoDet -> N RelClause
npNodet__n_relclause = GrammarRule(
    SourceSpec(POSType.NP_NoDet),
    [
        ExpansionSpec(
            POSType.Noun,
            {
                AttributeType.ANIMACY: INHERIT,
                AttributeType.CASE: INHERIT,
                AttributeType.NUMBER: INHERIT,
            },
        ),
        ExpansionSpec(
            POSType.RelClause,
            {
                AttributeType.ANIMACY: INHERIT,
            },
        ),
    ],
    weight=0.2,
)

npNodet_expansions = [
    npNodet__n,
    np_nodet__adjp_np_nodet,
    npNodet__adjp_n,
    npNodet__n_relclause,
]



# NP -> Det NP_NoDet
from greenideas.attributes.attribute_type import AttributeType
from greenideas.attributes.person import Person
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.expansion_spec import INHERIT, ExpansionSpec
from greenideas.rules.grammar_rule import GrammarRule
from greenideas.rules.source_spec import SourceSpec

np__det_npNodet = GrammarRule(
    SourceSpec(
        POSType.NP,
        {AttributeType.PERSON: Person.THIRD},
    ),
    [
        ExpansionSpec(
            POSType.Det,
            {
                AttributeType.CASE: INHERIT,
                AttributeType.NUMBER: INHERIT,
            },
        ),
        ExpansionSpec(
            POSType.NP_NoDet,
            {
                AttributeType.ANIMACY: INHERIT,
                AttributeType.CASE: INHERIT,
                AttributeType.NUMBER: INHERIT,
            },
        ),
    ],
)

# NP -> Pron
np__pron = GrammarRule(
    SourceSpec(POSType.NP),
    [
        ExpansionSpec(
            POSType.Pron,
            {
                AttributeType.ANIMACY: INHERIT,
                AttributeType.NUMBER: INHERIT,
                AttributeType.PERSON: INHERIT,
                AttributeType.CASE: INHERIT,
            },
        )
    ],
    weight=0.2,
)

np_expansions = [
    np__det_npNodet,
    np__pron,
]



class RuleNotFoundError(Exception):
    pass


class RelevantAttributesNotSpecified(Exception):
    pass


class InvalidGrammarRule(Exception):
    pass


class TwaddleConversionError(Exception):
    pass



from typing import Any, Optional

from greenideas.attributes.attribute_type import AttributeType
from greenideas.exceptions import InvalidGrammarRule
from greenideas.parts_of_speech.pos_type_attributes import POSTYPE_ATTRIBUTE_MAP
from greenideas.parts_of_speech.pos_types import POSType


class SourceSpec:
    def __init__(
        self, pos_type: POSType, attribute_constraints: dict[AttributeType, Any] = None
    ):
        self.pos_type = pos_type
        self.attribute_constraints = attribute_constraints or {}
        for attr, val in self.attribute_constraints.items():
            if attr not in POSTYPE_ATTRIBUTE_MAP[pos_type]:
                raise InvalidGrammarRule(
                    f"SourceSpec includes constraints on {str(attr)}, not defined as relevant attribute for {pos_type}"
                )
            if not isinstance(val, attr.value_type):
                if isinstance(val, list) and all(
                    isinstance(v, attr.value_type) for v in val
                ):
                    continue
                raise TypeError(
                    f"Value for {attr.name} must be of type {str(attr)} (or list of), got {type(val)}"
                )

    def get_constraint(self, attr_type: AttributeType) -> Optional[dict]:
        return self.attribute_constraints.get(attr_type, None)

    def __repr__(self):
        return f"SourceSpec({self.pos_type}, {self.attribute_constraints})"



from greenideas.attributes.attribute_type import AttributeType
from greenideas.attributes.number import Number
from greenideas.exceptions import TwaddleConversionError
from greenideas.parts_of_speech.pos_node import POSNode
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.twaddle.twaddle_tag import build_twaddle_tag


class DetFormattingHandler:
    @staticmethod
    def format(node: POSNode) -> str:
        if node.type != POSType.Det:
            raise TwaddleConversionError(
                f"Tried to use DetFormattingHandler on {node.type}"
            )
        name = "det"
        number = node.attributes.get(AttributeType.NUMBER)
        form = "pl" if number == Number.PLURAL else "sg"
        return build_twaddle_tag(name, form=form)



from greenideas.attributes.animacy import Animacy
from greenideas.attributes.attribute_type import AttributeType
from greenideas.attributes.case import Case
from greenideas.attributes.number import Number
from greenideas.exceptions import TwaddleConversionError
from greenideas.parts_of_speech.pos_node import POSNode
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.twaddle.twaddle_tag import build_twaddle_tag


class NounFormattingHandler:
    @staticmethod
    def format(node: POSNode) -> str:
        if node.type != POSType.Noun:
            raise TwaddleConversionError(
                f"Tried to use NounFormattingHandler on {node.type}"
            )
        name = "noun"
        class_specifier = None
        animacy = node.attributes.get(AttributeType.ANIMACY)
        number = node.attributes.get(AttributeType.NUMBER)
        case = node.attributes.get(AttributeType.CASE)
        form = "pl" if number == Number.PLURAL else "sg"
        if case == Case.GENITIVE:
            form += "gen"
        match animacy:
            case Animacy.ANIMATE:
                class_specifier = "animate"
            case Animacy.INANIMATE:
                class_specifier = "inanimate"
        return build_twaddle_tag(name, class_specifier=class_specifier, form=form)



import logging
import readline  # noqa: F401
import sys
from importlib.resources import files

from twaddle.runner import TwaddleRunner

from greenideas.grammar_engine import GrammarEngine
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.default_english_rules.default_rules import default_rules
from greenideas.twaddle.default_formatters.default_formatting_handlers import (
    default_formatting_handlers,
)
from greenideas.twaddle.twaddle_formatter import TwaddleFormatter


def main():
    logging.basicConfig(filename="greenideas.log", level=logging.INFO)
    if len(sys.argv) < 2:
        dictionary_path = files("greenideas.default_dictionary")
    else:
        dictionary_path = sys.argv[1]
    twaddle_runner = TwaddleRunner(dictionary_path)

    engine = GrammarEngine()
    engine.add_ruleset(default_rules)
    tree = engine.generate_tree(POSType.S)
    print(tree)

    formatter = TwaddleFormatter()
    for type, handler in default_formatting_handlers.items():
        formatter.register_formatting_handler(type, handler)

    twaddle_string = formatter.format_as_sentence(tree)
    print(twaddle_string)
    print(twaddle_runner.run_sentence(twaddle_string))


if __name__ == "__main__":
    main()



from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Self

from greenideas.attributes.attribute_type import AttributeType


@dataclass
class AttributeSet:
    _values: Dict[AttributeType, Any] = field(default_factory=dict)

    def get(self, attr: AttributeType) -> Optional[Any]:
        return self._values.get(attr)

    def set(self, attr: AttributeType, value: Any):
        if not isinstance(value, attr.value_type):
            raise TypeError(
                f"Value for {attr.name} must be of type {attr.value_type.__name__}, got {type(value).__name__}"
            )
        self._values[attr] = value

    def merge(self, other: Self, overwrite: bool = False) -> Self:
        result = AttributeSet()
        for attr, value in self._values.items():
            result._values[attr] = value
        for attr, value in other._values.items():
            if overwrite or attr not in result._values:
                result._values[attr] = value
        return result

    def __contains__(self, attr: AttributeType) -> bool:
        return attr in self._values



from greenideas.attributes.attribute_type import AttributeType
from greenideas.exceptions import InvalidGrammarRule
from greenideas.parts_of_speech.pos_node import POSNode
from greenideas.parts_of_speech.pos_type_attributes import relevant_attributes
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.expansion_spec import INHERIT, ExpansionSpec


class GrammarRule:
    def __init__(
        self,
        source: ExpansionSpec,
        expansion: list[ExpansionSpec],
        weight: float = 1.0,
    ):
        self.source = source
        self.pos = source.pos_type
        self.source_constraints = source.attribute_constraints
        self.expansion = expansion
        self.weight = weight

    def validate_attribute_constraint(
        self, pos_type: POSType, attr: AttributeType, value
    ) -> bool:
        if value == INHERIT and attr not in relevant_attributes(pos_type):
            raise InvalidGrammarRule(
                f"Rule {str(self)} requires inheriting {attr} from {pos_type} "
                f"but {attr} is not a relevant attribute for {pos_type}"
            )

    def get_child_spec(self, idx: int) -> ExpansionSpec:
        return self.expansion[idx]

    def __repr__(self):
        sep = "\n\n"
        return f"{self.pos}({self.source_constraints}) -> [{sep.join(str(item) for item in self.expansion)}]"

    def is_applicable_to_node(self, node: POSNode) -> bool:
        for attr_type, constraint in self.source_constraints.items():
            if isinstance(constraint, list):
                if node.attributes.get(attr_type) not in constraint:
                    return False
            elif node.attributes.get(attr_type) != constraint:
                return False
        return True



from typing import Optional


def build_twaddle_tag(
    dict_name: str,
    class_specifier: Optional[str | list[str]] = None,
    form: Optional[str] = None,
) -> str:
    tag_contents = dict_name
    if class_specifier:
        if isinstance(class_specifier, list):
            class_specifier = "-".join(class_specifier)
        tag_contents += f"-{class_specifier}"
    if form:
        tag_contents += f".{form}"
    return f"<{tag_contents}>"



from greenideas.attributes.attribute_type import AttributeType
from greenideas.attributes.valency import Valency
from greenideas.exceptions import TwaddleConversionError
from greenideas.parts_of_speech.pos_node import POSNode
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.twaddle.twaddle_tag import build_twaddle_tag


class VerbBareFormattingHandler:
    @staticmethod
    def format(node: POSNode) -> str:
        if node.type != POSType.Verb_Bare:
            raise TwaddleConversionError(
                f"Tried to use VerbBareFormattingHandler on {node.type}"
            )

        valency = node.attributes.get(AttributeType.VALENCY)
        match valency:
            case Valency.MONOVALENT:
                class_specifier = "monovalent"
            case Valency.DIVALENT:
                class_specifier = "divalent"
            case _:
                raise TwaddleConversionError(f"Invalid valency: {valency}")
        return build_twaddle_tag("verb", class_specifier=class_specifier)



from enum import auto

from greenideas.attributes.grammatical_attribute import GrammaticalAttribute


class Valency(GrammaticalAttribute):
    MONOVALENT = auto()
    DIVALENT = auto()
    # TRIVALENT = auto()  # handling trivalent verbs will require twaddle labels
    # twaddle labels will be handled as part of forthcoming pronoun update

    def __str__(self) -> str:
        return self.name.lower()



from greenideas.attributes.attribute_type import AttributeType
from greenideas.attributes.number import Number
from greenideas.attributes.person import Person
from greenideas.attributes.tense import Tense
from greenideas.exceptions import TwaddleConversionError
from greenideas.parts_of_speech.pos_node import POSNode
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.twaddle.twaddle_tag import build_twaddle_tag


class AuxDoFormattingHandler:
    @staticmethod
    def format(node: POSNode) -> str:
        if node.type != POSType.Aux_do:
            raise TwaddleConversionError(
                f"Tried to use AuxDoFormattingHandler on {node.type}"
            )
        name = "aux"
        form = None
        number = node.attributes.get(AttributeType.NUMBER)
        person = node.attributes.get(AttributeType.PERSON)
        tense = node.attributes.get(AttributeType.TENSE)
        if tense == Tense.PAST:
            form = "past"
        elif person == Person.THIRD and number == Number.SINGULAR:
            form = "s"
        return build_twaddle_tag(name, class_specifier="do", form=form)



from greenideas.exceptions import TwaddleConversionError
from greenideas.parts_of_speech.pos_node import POSNode
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.twaddle.twaddle_tag import build_twaddle_tag


class CoordconjFormattingHandler:
    @staticmethod
    def format(node: POSNode) -> str:
        if node.type != POSType.CoordConj:
            raise TwaddleConversionError(
                f"Tried to use CoordconjFormattingHandler on {node.type}"
            )
        return build_twaddle_tag("conj", class_specifier="coordinating")



from typing import Any, Optional

from greenideas.attributes.attribute_type import AttributeType
from greenideas.exceptions import InvalidGrammarRule
from greenideas.parts_of_speech.pos_type_attributes import relevant_attributes
from greenideas.parts_of_speech.pos_types import POSType


class InheritSentinel:
    def __repr__(self):
        return "INHERIT"


INHERIT = InheritSentinel()


class ExpansionSpec:
    def __init__(
        self, pos_type: POSType, attribute_constraints: dict[AttributeType, Any] = None
    ):
        self.pos_type = pos_type
        self.attribute_constraints = attribute_constraints or {}
        for attr in self.attribute_constraints:
            if attr not in relevant_attributes(self.pos_type):
                raise InvalidGrammarRule(
                    f"Attribute constraints on {attr}, not defined as relevant attribute for {pos_type}"
                )

    def get_constraint(self, attr_type: AttributeType) -> Optional[dict]:
        return self.attribute_constraints.get(attr_type, None)

    def __repr__(self):
        return f"ExpansionSpec({self.pos_type}, {self.attribute_constraints})"

    def __str__(self):
        return f"ExpansionSpec(pos_type={self.pos_type}, attribute_constraints={self.attribute_constraints})"



from enum import Enum, auto


class POSType(Enum):
    S = auto()

    AdjP = auto()
    AdvP = auto()
    AuxP = auto()
    ModalP = auto()
    NP = auto()
    NP_NoDet = auto()
    PP = auto()
    RelClause = auto()
    VP = auto()
    VP_AfterModal = auto()
    VP_Bare = auto()

    Adj = auto()
    Adv = auto()
    Aux_do = auto()  # do support
    Aux_finite = auto()  # tensed auxiliaries, have/be
    Be = auto()
    CoordConj = auto()
    Det = auto()
    Modal = auto()
    Noun = auto()
    Prep = auto()
    Pron = auto()
    RelativePron = auto()
    Subordinator = auto()
    Verb = auto()
    Verb_AfterModal = auto()
    Verb_Bare = auto()



from greenideas.attributes.aspect import Aspect
from greenideas.attributes.attribute_type import AttributeType
from greenideas.attributes.tense import Tense
from greenideas.exceptions import TwaddleConversionError
from greenideas.parts_of_speech.pos_node import POSNode
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.twaddle.twaddle_tag import build_twaddle_tag


class ModalFormattingHandler:
    @staticmethod
    def format(node: POSNode) -> str:
        if node.type != POSType.Modal:
            raise TwaddleConversionError(
                f"Tried to use ModalFormattingHandler on {node.type}"
            )
        name = "modal"
        tense = node.attributes.get(AttributeType.TENSE)
        aspect = node.attributes.get(AttributeType.ASPECT)
        match tense:
            case Tense.PAST:
                form = "past"
            case Tense.PRESENT:
                form = "pres"
            case _:
                raise TwaddleConversionError(
                    f"Invalid tense {tense} for ModalFormattingHandler"
                )
        match aspect:
            case Aspect.PERFECT:
                form += "perf"
            case Aspect.PROGRESSIVE:
                form += "prog"
            case Aspect.PERFECT_PROGRESSIVE:
                form += "perfprog"
            case Aspect.SIMPLE:
                pass
            case _:
                raise TwaddleConversionError(
                    f"Invalid aspect {aspect} for ModalFormattingHandler"
                )
        return build_twaddle_tag(name, form=form)



import logging
import random

from greenideas.attributes.grammatical_attribute import GrammaticalAttribute
from greenideas.exceptions import RuleNotFoundError
from greenideas.grammar import Grammar
from greenideas.parts_of_speech.pos_node import POSNode
from greenideas.parts_of_speech.pos_type_attributes import relevant_attributes
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.expansion_spec import INHERIT, ExpansionSpec
from greenideas.rules.grammar_rule import GrammarRule
from greenideas.rules.grammar_ruleset import GrammarRuleset

logger = logging.getLogger(__name__)


class GrammarEngine:
    def __init__(self):
        self.grammar = Grammar()

    def add_rule(self, rule: GrammarRule):
        self.grammar.add_rule(rule)

    def add_ruleset(self, ruleset: GrammarRuleset):
        for rule in ruleset.rules:
            self.add_rule(rule)

    def clear_rules(self):
        self.grammar.clear_rules()

    def generate_tree(self, start: POSNode | POSType) -> POSNode:
        if isinstance(start, POSNode):
            node = start
            self._assign_random_attributes(node)
        elif isinstance(start, POSType):
            node = POSNode(type=start)
            self._assign_random_attributes(node)
        else:
            raise ValueError("start must be a POSType or POSNode")
        if len(self.grammar.get_applicable_rules(node)) == 0:
            raise RuleNotFoundError(f"No rule found to expand type {node.type}")
        logger.info(node.attributes)
        return self._expand_to_tree(node)

    def _expand_to_tree(self, node: POSNode) -> POSNode:
        rules = self.grammar.get_applicable_rules(node)
        if not rules:
            return node
        rule = random.choices(rules, weights=[r.weight for r in rules])[0]
        logger.info(f"expanding {node.type} with rule {rule}\n\n\n")
        children = []
        for _, spec in enumerate(rule.expansion):
            if isinstance(spec, ExpansionSpec):
                child = POSNode(type=spec.pos_type)
                for attr_type, constraint in spec.attribute_constraints.items():
                    if constraint is not None:
                        if constraint == INHERIT:
                            child.attributes.set(
                                attr_type, node.attributes.get(attr_type)
                            )
                        elif isinstance(constraint, list):
                            child.attributes.set(
                                attr_type, GrammaticalAttribute.random_from(constraint)
                            )
                        else:
                            child.attributes.set(attr_type, constraint)
                self._assign_random_attributes(child)
                children.append(self._expand_to_tree(child))
        node.children = children
        return node

    def _assign_random_attributes(self, node: POSNode) -> None:
        for attr_type in relevant_attributes(node.type):
            if attr_type not in node.attributes:
                possible_values = list(attr_type.value_type)
                node.attributes.set(
                    attr_type, GrammaticalAttribute.random_from(possible_values)
                )



from greenideas.rules.grammar_rule import GrammarRule


class GrammarRuleset:
    def __init__(self):
        self.rules = list()

    def add(self, rule: GrammarRule):
        self.rules.append(rule)

    def add_rules(self, rules: list[GrammarRule]):
        for rule in rules:
            self.add(rule)



from greenideas.parts_of_speech.pos_types import POSType
from greenideas.twaddle.default_formatters.adj_formatting_handler import (
    AdjFormattingHandler,
)
from greenideas.twaddle.default_formatters.adv_formatting_handler import (
    AdvFormattingHandler,
)
from greenideas.twaddle.default_formatters.aux_do_formatting_handler import (
    AuxDoFormattingHandler,
)
from greenideas.twaddle.default_formatters.aux_finite_formatting_handler import (
    AuxFiniteFormattingHandler,
)
from greenideas.twaddle.default_formatters.be_formatting_handler import (
    BeFormattingHandler,
)
from greenideas.twaddle.default_formatters.coordconj_formatting_handler import (
    CoordconjFormattingHandler,
)
from greenideas.twaddle.default_formatters.det_formatting_handler import (
    DetFormattingHandler,
)
from greenideas.twaddle.default_formatters.modal_formatting_handler import (
    ModalFormattingHandler,
)
from greenideas.twaddle.default_formatters.noun_formatting_handler import (
    NounFormattingHandler,
)
from greenideas.twaddle.default_formatters.prep_formatting_handler import (
    PrepFormattingHandler,
)
from greenideas.twaddle.default_formatters.pron_formatting_handler import (
    PronFormattingHandler,
)
from greenideas.twaddle.default_formatters.relative_pronoun_formatting_handler import (
    RelativePronounFormattingHandler,
)
from greenideas.twaddle.default_formatters.subordinator_formatting_handler import (
    SubordinatorFormattingHandler,
)
from greenideas.twaddle.default_formatters.verb_after_modal_formatting_handler import (
    VerbAfterModalFormattingHandler,
)
from greenideas.twaddle.default_formatters.verb_bare_formatting_handler import (
    VerbBareFormattingHandler,
)
from greenideas.twaddle.default_formatters.verb_formatting_handler import (
    VerbFormattingHandler,
)
from greenideas.twaddle.twaddle_formatting_handler import TwaddleFormattingHandler

default_formatting_handlers: dict[POSType, TwaddleFormattingHandler] = {
    POSType.Adj: AdjFormattingHandler,
    POSType.Adv: AdvFormattingHandler,
    POSType.Aux_do: AuxDoFormattingHandler,
    POSType.Aux_finite: AuxFiniteFormattingHandler,
    POSType.Be: BeFormattingHandler,
    POSType.CoordConj: CoordconjFormattingHandler,
    POSType.Det: DetFormattingHandler,
    POSType.Modal: ModalFormattingHandler,
    POSType.Noun: NounFormattingHandler,
    POSType.Prep: PrepFormattingHandler,
    POSType.Pron: PronFormattingHandler,
    POSType.RelativePron: RelativePronounFormattingHandler,
    POSType.Subordinator: SubordinatorFormattingHandler,
    POSType.Verb: VerbFormattingHandler,
    POSType.Verb_AfterModal: VerbAfterModalFormattingHandler,
    POSType.Verb_Bare: VerbBareFormattingHandler,
}



from enum import Enum
from typing import Type

from greenideas.attributes.animacy import Animacy
from greenideas.attributes.aspect import Aspect
from greenideas.attributes.case import Case
from greenideas.attributes.number import Number
from greenideas.attributes.person import Person
from greenideas.attributes.tense import Tense
from greenideas.attributes.valency import Valency


class AttributeType(Enum):
    ASPECT = ("aspect", Aspect)
    ANIMACY = ("animacy", Animacy)
    CASE = ("case", Case)
    NUMBER = ("number", Number)
    PERSON = ("person", Person)
    TENSE = ("tense", Tense)
    VALENCY = ("valency", Valency)

    def __init__(self, attr_name: str, value_type: Type):
        self.attr_name = attr_name
        self.value_type = value_type

    def __str__(self) -> str:
        return self.attr_name



from enum import auto

from greenideas.attributes.grammatical_attribute import GrammaticalAttribute


class Animacy(GrammaticalAttribute):
    ANIMATE = auto()
    INANIMATE = auto()

    def __str__(self) -> str:
        return self.name.lower()



from greenideas.attributes.aspect import Aspect
from greenideas.attributes.attribute_type import AttributeType
from greenideas.attributes.number import Number
from greenideas.attributes.person import Person
from greenideas.attributes.tense import Tense
from greenideas.exceptions import TwaddleConversionError
from greenideas.parts_of_speech.pos_node import POSNode
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.twaddle.twaddle_tag import build_twaddle_tag


class BeFormattingHandler:
    @staticmethod
    def format(node: POSNode) -> str:
        if node.type != POSType.Be:
            raise TwaddleConversionError(
                f"Tried to use BeFormattingHandler on {node.type}"
            )
        name = "be"
        form = None
        aspect = node.attributes.get(AttributeType.ASPECT)
        number = node.attributes.get(AttributeType.NUMBER)
        person = node.attributes.get(AttributeType.PERSON)
        tense = node.attributes.get(AttributeType.TENSE)
        if aspect == Aspect.PROGRESSIVE:
            form = "gerund"
        elif aspect == Aspect.PERFECT:
            form = "pastpart"
        else:
            if number == Number.SINGULAR:
                if person == Person.FIRST:
                    form = "1sg"
                elif person == Person.THIRD:
                    form = "3sg"
                else:
                    form = "other"
            else:
                form = "other"
            if tense == Tense.PRESENT:
                form += "pres"
            else:
                form += "past"
        return build_twaddle_tag(name, form=form)



from greenideas.attributes.aspect import Aspect
from greenideas.attributes.attribute_type import AttributeType
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.expansion_spec import INHERIT, ExpansionSpec
from greenideas.rules.grammar_rule import GrammarRule
from greenideas.rules.source_spec import SourceSpec

relC__relPron_VP = GrammarRule(
    SourceSpec(POSType.RelClause),
    [
        ExpansionSpec(POSType.RelativePron, {AttributeType.ANIMACY: INHERIT}),
        ExpansionSpec(
            POSType.VP,
            {
                AttributeType.ASPECT: [Aspect.SIMPLE, Aspect.PERFECT],
            },
        ),
    ],
)

relC_expansions = [
    relC__relPron_VP,
]



from greenideas.parts_of_speech.pos_node import POSNode
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.grammar_rule import GrammarRule


class Grammar:
    def __init__(self):
        self.rules = {}

    def add_rule(self, rule: GrammarRule):
        part_of_speech = rule.pos
        if part_of_speech in self.rules:
            self.rules[part_of_speech].append(rule)
        else:
            self.rules[part_of_speech] = [rule]

    def clear_rules(self):
        self.rules = {}

    def get_rules(self, part_of_speech: POSType) -> list[GrammarRule]:
        return self.rules.get(part_of_speech, [])

    def get_applicable_rules(self, node: POSNode) -> list[GrammarRule]:
        candidates = self.get_rules(node.type)
        rules = [
            candidate
            for candidate in candidates
            if candidate.is_applicable_to_node(node)
        ]
        return rules

    def has_expansion(self, part_of_speech: POSType) -> bool:
        return part_of_speech in self.rules



from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.expansion_spec import ExpansionSpec
from greenideas.rules.grammar_rule import GrammarRule
from greenideas.rules.source_spec import SourceSpec

# VP_Bare -> Adv VP(bare)
vpBare__adv_vpBare = GrammarRule(
    SourceSpec(POSType.VP_Bare),
    [
        ExpansionSpec(POSType.Adv),
        ExpansionSpec(
            POSType.VP_Bare,
        ),
    ],
    weight=0.2,
)

# VP_Bare -> Verb(bare)
vpBare__vBare = GrammarRule(
    SourceSpec(POSType.VP_Bare),
    [
        ExpansionSpec(
            POSType.Verb_Bare,
        )
    ],
)

vpBare_expansions = [vpBare__adv_vpBare, vpBare__vBare]



from greenideas.rules.default_english_rules.adjP_expansions import adjP_expansions
from greenideas.rules.default_english_rules.advP_expansions import advP_expansions
from greenideas.rules.default_english_rules.auxP_expansions import auxP_expansions
from greenideas.rules.default_english_rules.modalP_expansions import modalP_expansions
from greenideas.rules.default_english_rules.np_expansions import np_expansions
from greenideas.rules.default_english_rules.npNodet_expansions import npNodet_expansions
from greenideas.rules.default_english_rules.pp_expansions import pp_expansions
from greenideas.rules.default_english_rules.relClause_expansions import relC_expansions
from greenideas.rules.default_english_rules.s_expansions import s_expansions
from greenideas.rules.default_english_rules.vp_expansions import vp_expansions
from greenideas.rules.default_english_rules.vpAfterModal_expansions import (
    vpAfterModal_expansions,
)
from greenideas.rules.default_english_rules.vpBare_expansions import vpBare_expansions
from greenideas.rules.grammar_ruleset import GrammarRuleset

default_rules = GrammarRuleset()


default_rules.add_rules(s_expansions)
default_rules.add_rules(adjP_expansions)
default_rules.add_rules(advP_expansions)
default_rules.add_rules(auxP_expansions)
default_rules.add_rules(modalP_expansions)
default_rules.add_rules(np_expansions)
default_rules.add_rules(npNodet_expansions)
default_rules.add_rules(pp_expansions)
default_rules.add_rules(relC_expansions)
default_rules.add_rules(vp_expansions)
default_rules.add_rules(vpAfterModal_expansions)
default_rules.add_rules(vpBare_expansions)



from greenideas.attributes.attribute_type import AttributeType
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.expansion_spec import INHERIT, ExpansionSpec
from greenideas.rules.grammar_rule import GrammarRule
from greenideas.rules.source_spec import SourceSpec

# VPAfterModal -> Adv VAfterModal
vpAfterModal__adv_vAfterModal = GrammarRule(
    SourceSpec(POSType.VP_AfterModal),
    [
        ExpansionSpec(POSType.Adv),
        ExpansionSpec(
            POSType.Verb_AfterModal,
            {
                AttributeType.ASPECT: INHERIT,
            },
        ),
    ],
)

# VPAfterModal -> VAfterModal
vpAfterModal__vAfterModal = GrammarRule(
    SourceSpec(POSType.VP_AfterModal),
    [
        ExpansionSpec(
            POSType.Verb_AfterModal,
            {
                AttributeType.ASPECT: INHERIT,
            },
        ),
    ],
)

# VPAfterModal -> VAfterModal AdvP
vpAfterModal__vAfterModal_advP = GrammarRule(
    SourceSpec(POSType.VP_AfterModal),
    [
        ExpansionSpec(
            POSType.Verb_AfterModal,
            {
                AttributeType.ASPECT: INHERIT,
            },
        ),
        ExpansionSpec(POSType.AdvP),
    ],
)

vpAfterModal_expansions = [
    vpAfterModal__adv_vAfterModal,
    vpAfterModal__vAfterModal,
    vpAfterModal__vAfterModal_advP,
]



from enum import Enum, unique
from random import choices
from typing import Self


@unique
class GrammaticalAttribute(Enum):

    def __init__(self, label, weight=1):
        self.label = label
        self.weight = weight

    @classmethod
    def random_from(cls, allowed: list[Self]) -> Self:
        weights = [elem.weight for elem in allowed]
        return choices(allowed, weights=weights)[0]



from typing import Protocol

from greenideas.exceptions import TwaddleConversionError
from greenideas.parts_of_speech.pos_node import POSNode


class TwaddleFormattingHandler(Protocol):
    @staticmethod
    def format(node: POSNode) -> str:
        raise TwaddleConversionError(
            "Twaddle Formatter defines a protocol and is not " "directly usable"
        )



from enum import auto

from greenideas.attributes.grammatical_attribute import GrammaticalAttribute


class Case(GrammaticalAttribute):
    NOMINATIVE = auto()
    OBJECTIVE = auto()
    GENITIVE = auto()



from greenideas.parts_of_speech.pos_types import POSType
from greenideas.rules.expansion_spec import ExpansionSpec
from greenideas.rules.grammar_rule import GrammarRule
from greenideas.rules.source_spec import SourceSpec

# AdvP -> Adv
advp__adv = GrammarRule(
    SourceSpec(POSType.AdvP),
    [ExpansionSpec(POSType.Adv)],
)

# AdvP -> Adv conj Adv
advp__adv_conj_adv = GrammarRule(
    SourceSpec(POSType.AdvP),
    [
        ExpansionSpec(POSType.Adv),
        ExpansionSpec(POSType.CoordConj),
        ExpansionSpec(POSType.Adv),
    ],
    weight=0.2,
)

# AdvP -> PP
advp__pp = GrammarRule(
    SourceSpec(POSType.AdvP),
    [ExpansionSpec(POSType.PP)],
)

advP_expansions = [
    advp__adv,
    advp__adv_conj_adv,
    advp__pp,
]



from greenideas.attributes.animacy import Animacy
from greenideas.attributes.attribute_type import AttributeType
from greenideas.parts_of_speech.pos_node import POSNode
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.twaddle.twaddle_tag import build_twaddle_tag


class RelativePronounFormattingHandler:
    @staticmethod
    def format(node: POSNode) -> str:
        if node.type != POSType.RelativePron:
            raise ValueError(
                f"Tried to use RelativePronounFormattingHandler on {node.type}"
            )
        name = "rel"
        animacy = node.attributes.get(AttributeType.ANIMACY)
        match animacy:
            case Animacy.ANIMATE:
                class_specifier = "animate"
            case Animacy.INANIMATE:
                class_specifier = "inanimate"
        return build_twaddle_tag(name, class_specifier=class_specifier)



from greenideas.attributes.animacy import Animacy
from greenideas.attributes.attribute_type import AttributeType
from greenideas.attributes.case import Case
from greenideas.attributes.number import Number
from greenideas.attributes.person import Person
from greenideas.exceptions import TwaddleConversionError
from greenideas.parts_of_speech.pos_node import POSNode
from greenideas.parts_of_speech.pos_types import POSType
from greenideas.twaddle.twaddle_tag import build_twaddle_tag


class PronFormattingHandler:
    @staticmethod
    def format(node: POSNode) -> str:
        if node.type != POSType.Pron:
            raise TwaddleConversionError(
                f"Tried to use PronFormattingHandler on {node.type}"
            )
        name = "pron"
        class_specifiers = list()
        person = node.attributes.get(AttributeType.PERSON)
        animacy = node.attributes.get(AttributeType.ANIMACY)
        match person:
            case Person.FIRST:
                class_specifiers.append("firstperson")
            case Person.SECOND:
                class_specifiers.append("secondperson")
            case Person.THIRD:
                class_specifiers.append("thirdperson")
                # animacy only relevant in third person
                match animacy:
                    case Animacy.ANIMATE:
                        class_specifiers.append("animate")
                    case Animacy.INANIMATE:
                        class_specifiers.append("inanimate")
        number = node.attributes.get(AttributeType.NUMBER)
        case = node.attributes.get(AttributeType.CASE)
        form = "pl" if number == Number.PLURAL else "sg"
        if case == Case.GENITIVE:
            form += "gen"
        elif case == Case.OBJECTIVE:
            form += "obj"
        return build_twaddle_tag(name, class_specifier=class_specifiers, form=form)



from greenideas.attributes.attribute_type import AttributeType
from greenideas.parts_of_speech.pos_types import POSType

POSTYPE_ATTRIBUTE_MAP = {
    POSType.S: {
        AttributeType.ANIMACY,
        AttributeType.NUMBER,
        AttributeType.PERSON,
        AttributeType.TENSE,
    },
    POSType.AdjP: set(),
    POSType.AdvP: set(),
    POSType.AuxP: {
        AttributeType.ASPECT,
        AttributeType.NUMBER,
        AttributeType.PERSON,
        AttributeType.TENSE,
    },
    POSType.Be: {
        AttributeType.ASPECT,
        AttributeType.NUMBER,
        AttributeType.PERSON,
        AttributeType.TENSE,
    },
    POSType.ModalP: {
        AttributeType.ASPECT,
        AttributeType.TENSE,
    },
    POSType.NP: {
        AttributeType.ANIMACY,
        AttributeType.CASE,
        AttributeType.NUMBER,
        AttributeType.PERSON,
    },
    POSType.NP_NoDet: {
        AttributeType.ANIMACY,
        AttributeType.CASE,
        AttributeType.NUMBER,
        AttributeType.PERSON,
    },
    POSType.PP: {},
    POSType.RelClause: {
        AttributeType.ANIMACY,
    },
    POSType.VP: {
        AttributeType.ANIMACY,
        AttributeType.ASPECT,
        AttributeType.NUMBER,
        AttributeType.PERSON,
        AttributeType.TENSE,
        AttributeType.VALENCY,
    },
    POSType.VP_AfterModal: {
        AttributeType.ASPECT,
        AttributeType.VALENCY,
    },
    POSType.VP_Bare: {
        AttributeType.VALENCY,
    },
    POSType.Adj: set(),
    POSType.Adv: set(),
    POSType.Aux_do: {
        AttributeType.TENSE,
        AttributeType.NUMBER,
        AttributeType.PERSON,
        AttributeType.ASPECT,
    },
    POSType.Aux_finite: {
        AttributeType.TENSE,
        AttributeType.NUMBER,
        AttributeType.PERSON,
        AttributeType.ASPECT,
    },
    POSType.CoordConj: set(),
    POSType.Det: {
        AttributeType.CASE,
        AttributeType.NUMBER,
    },
    POSType.Modal: {
        AttributeType.ASPECT,
        AttributeType.TENSE,
    },
    POSType.Noun: {
        AttributeType.ANIMACY,
        AttributeType.CASE,
        AttributeType.NUMBER,
    },
    POSType.Prep: set(),
    POSType.Pron: {
        AttributeType.ANIMACY,
        AttributeType.CASE,
        AttributeType.NUMBER,
        AttributeType.PERSON,
    },
    POSType.RelativePron: {
        AttributeType.ANIMACY,
    },
    POSType.Subordinator: set(),
    POSType.Verb: {
        AttributeType.ANIMACY,
        AttributeType.ASPECT,
        AttributeType.NUMBER,
        AttributeType.PERSON,
        AttributeType.TENSE,
        AttributeType.VALENCY,
    },
    POSType.Verb_AfterModal: {
        AttributeType.ASPECT,
        AttributeType.VALENCY,
    },
    POSType.Verb_Bare: {AttributeType.VALENCY},
}


def relevant_attributes(pos_type: POSType) -> set[AttributeType]:
    if pos_type not in POSTYPE_ATTRIBUTE_MAP:
        raise ValueError(f"No relevant attributes specified for POSType: {pos_type}")
    return POSTYPE_ATTRIBUTE_MAP[pos_type]
